// This may look like C code, but it is really -*- C++ -*-
//  
// SoundAF.h
// 
// C++ interface to soundfiles
// based on SGI calls
// dpwe@interval.com 1994aug03
//

#ifndef _SOUNDAF_H
#define _SOUNDAF_H

#include <audiofile.h>

typedef unsigned char byte;

enum  Format {
    Sfmt_SHORT = 2,
    Sfmt_LONG  = 4,
    Sfmt_FLOAT = 36,
};

int SfmtSize(int format);
    /* return the bytes per sample of a particular data format */

class Sound {
private:
// Details of the source or destination of the bits
    char		*filename;	// name of the attached soundfile
    AFfilehandle	fileHandle;	// AF reference handle
    char		mode;		// 'r' for read, 'w' for write
    int			flags;		// IS_STDIO, IS_REALTIME etc
    long		pos;		// index of next frame to be accessed
// Details of the actual sound.  There are two aspects to this - 
//  what is actually in the file, and what is presented to the user.
//  The F* members describe the file, the U* describe the data passed 
//  through the read/write calls.
    int			Fchannels;	// 1 for mono, 2 for stereo
    Format		Fformat;	// i.e. INT/FLOAT, bits/sample
    double		FsampleRate;	// sample frames per second
    int			Uchannels;
    Format		Uformat;
    double		UsampleRate;
    long		numFrames;	// number of sample frames if known
// Buffer for reading stuff from disk
    byte		*buffer;	// some allocated space
    long		bufSize;	// number of bytes in buffer
    long		bufConts;	// count of meaningful bytes in buffer
// private functions
    long _readFrames(void *buf, long frames);	// file read without fmt conv.
    long _writeFrames(void *buf, long frames);	// file write without fmt conv.
    long checkBuffer(long bytesize);		// check buffer size

public:
// Constructors & destructor
    Sound(char modechr = 'r', 
	  double sampRate = 44100.0, int chans = 1, Format fmt = Sfmt_SHORT);
                                        // create a new sound object
    ~Sound(void);
// Accessors
    int getChannels(void)		{ return Uchannels; }
    int setChannels(int chans)		{ return Uchannels = chans; }
    Format getFormat(void)		{ return Uformat; }
    Format setFormat(Format fmt)	{ return Uformat = fmt; }
    double getSampleRate(void)		{ return UsampleRate; }
    double setSampleRate(double srate)	{ return UsampleRate = srate; }
    long getFrames(void)		{ return numFrames; }
    long setFrames(long frms)		{ return numFrames = frms; }
            // does setFrames make sense?  Not easily changed.
            // But could be handy for hinting the size of an output 
            // file to be written.
    char *getFilename(void)		{ return filename; }
    char getMode(void)			{ return mode; }
    char setMode(char modechr)		{  if(flags == 0 && 
					      (modechr == 'r'  
					       || modechr=='w')) mode=modechr;
					   return mode; }
// File operations	
    int open(char *filename);			// open file for input, 
                                                // or create for output.
                                                // Returns zero on success.
    long readFrames(void *buf, long frames);	// read sound data from file
    long writeFrames(void *buf, long frames);	// write sound data to file
    long frameTell(void);			// frame index of next to read
    long frameSeek(long frame);			// move file point to abs pos.
    int frameFeof(void);			// return nonzero if at EOF
};

#endif /* _SOUNDAF_H */
