static char rcsid[] = "$Header: /usr/people/dpwe/src/soundif/dat/lib/dat/RCS/dat.c,v 1.14 94/06/03 15:20:04 klee Exp Locker: dpwe $" ;

/*
 * dat.c
 *
 * $Source: /usr/people/dpwe/src/soundif/dat/lib/dat/RCS/dat.c,v $
 * $Date: 94/06/03 15:20:04 $
 * $Author: klee $
 * $Locker: dpwe $
 *
 * This file is part of the dat library.
 *
 * $Log:	dat.c,v $
 * Revision 1.14  94/06/03  15:20:04  klee
 * addeded ifdef for linux to d_flush
 * 
 * Revision 1.13  94/01/07  22:00:28  sbeck
 * Removed all evidence of floating point representation
 * since it is now completely handled by proper byte order
 * processing. Also changed usage of the predefined macro,
 * MACH_BYTE_ORDER to the function call mach_get_byte_order() 
 * which automatically determines the endian-ness of the host
 * machine.
 * 
 * Added handling of 'x' and 'X' datatypes.
 * 
 * Revision 1.12  1993/05/12  21:16:26  holtzman
 * Added support for Sunos 4.1.1.  In particular, added an
 * include of unistd.h and support for on_exit instead of atexit.
 *
 * Revision 1.11  1993/03/01  19:21:20  holtzman
 * On SGIs and possibly all SYSV machines, fopen with a mode of "a+" does
 * not allow writing to the pre-exiting portion of the file.  We need to
 * be able to do that.  So, opens now do a "a+" first, to create th
 * file if it doesn't exist, and then the file is immediately closed and
 * opened again with "r+".
 *
 * Revision 1.10  22/.1/.2  .0:.0:.2  holtzman
 * Removed some #includes
 * 
 * Revision 1.9  92/06/23  18:39:57  holtzman
 * changed a NULL to a 0 to eradicate a type check warning.
 * 
 * Revision 1.8  92/04/09  17:07:43  holtzman
 * float and byte rep mismatchs no longer exit program
 * 
 * Revision 1.7  92/02/20  11:36:54  klee
 * No longer needs to redefine exit() to close datfiles;
 * uses atexit() instead.
 * 
 * Revision 1.6  91/11/02  23:31:30  klee
 * updated for const declarations
 * 
 * Revision 1.5  91/11/02  22:34:45  klee
 * found lying around...
 * 
 * Revision 1.4  91/06/28  20:52:19  pizza
 * Fixed d_open to return NULL when it cannot change data directory.
 * 
 * Revision 1.9  90/07/19  10:27:39  pizza
 * Added support for float-representation and byte-order checking in
 * d_open.
 * 
 * Revision 1.8  89/01/13  02:42:18  jwang
 * Added M.I.T. copyright message.
 * 
 * Revision 1.7  88/01/21  23:12:45  jwang
 * Modify d_open() so that it returns NULL on more errors.
 * 
 * Revision 1.6  86/08/19  11:28:57  jwang
 * Replace fork by vfork if running on a vax.
 * 
 * Revision 1.5  86/07/13  17:44:46  jwang
 * In d_close, call _dclrinfo only if d!=NULL (i.e., if opened by
 * d_descr); otherwise, fileno(d_getfp(d)) is invalid.
 * 
 * Revision 1.4  86/06/12  15:19:16  jwang
 * Added #ifdef for BSD 4.3
 * 
 * Revision 1.3  86/03/08  12:29:56  steve
 * d_cleanup flushes rather than closes all files.
 * This is so that _daterrh can use it.
 * 
 * Revision 1.2  86/03/07  00:36:08  steve
 * Added call to _dclrinfo from d_close.
 * 
 * Revision 1.1  85/06/28  23:41:53  musicus
 * Initial revision
 * 
 *	History:
 * 		26-Oct-84 Added automatic closing of dat files.  (John Wang)
 *
 * A DATFILE struct has the following form:
 *    DATFILE
 *    {
 *    	struct d_token *db_h;		(ptr to token tree)
 *    	FILE *db_fp;			(FILE pointer for data
 *    	char *db_name;			(datfile name specified in call)
 *    	char db_flag;			(status flags such as _DB_WRHDR)
 *    	char db_nbs;			(log of number of bytes per sample)
 *    };
 * 	db_h normally points to a linked list of token structs, each pointing
 *	  to key and value strings.  If the header is empty or descriptor file
 *	  could not be found, db_h=0.  If the datfile is actually an ordinary
 *	  file and there can be no header, then db_h=-1.  ALL TOKEN STRUCTS
 *	  AND KEYWORD,VALUE STRINGS MUST BE MALLOC'ED FROM THE FREE LIST.
 *	  Flag bit _DB_WRHDR must be set if header is modified.
 *	db_fp is the file pointer for the opened data file.  If no data file
 *	  is open, db_fp=0
 *	db_name is the name of the dat file.  THIS MUST BE MALLOC'ED FROM THE
 *	  FREE LIST.
 *	db_flag has miscellaneous flags.
 *	  _DB_WRHDR indicates header modified in core.
 *	  _DB_GOTTYPE indicates db_nbs is meaningful.
 *	db_nbs is number places to shift sample count left to get byte count.
 *	  This is set by first d_read() or d_write().
 *
 */

/* Default file permissions for dat file components */
#ifndef DEFAULT_DIR_PERMS
#define DEFAULT_DIR_PERMS 0777
#endif

#ifndef DEFAULT_FILE_PERMS
#define DEFAULT_FILE_PERMS 0666
#endif

#define DEBUG 1
#ifdef DEBUG
#define debug(a,b)	fprintf(stderr,(a),(b))
#else
#define debug(a,b)
#endif

#include "copyright.h"
/* #include <stdio.h> */
/* #include <ctype.h> */
#include <sys/types.h>
#include <sys/stat.h>
#include "dat.h"

#include <sys/param.h>

#ifdef sun
#include <unistd.h>
#endif

#ifndef _NFILE
#define _NFILE NOFILE
#endif

#ifdef __sgi
#define vfork fork
#endif /* __sgi dpwe 1994dec06 */

char *calloc(), *realloc(), *malloc();
/* int free(); */

#define max(x, y)	(((x) < (y)) ? (y) : (x))
#define mabs(x)		((x < 0) ? -(x) : (x))
#define min(x, y)	(((x) < (y)) ? (x) : (y))

/* defines used by dat.c */
#define	DESCRIPTOR	"/descriptor" /* descriptor file name */
#define	DATA		"/data"	/* data file name */
#define	LTOKEN		'('	/* tokens for descriptor file. start of a */
#define	RTOKEN		')'	/* key entry, and end. */
#define	COMCHAR		';'	/* descriptor comment */
#define	ESCCHAR		'\\'	/* escape */

/* queue of pointers to open datfiles */
static DATFILE *d_iob[_NFILE];

/* pointer past end of queue */
static DATFILE **d_lastbuf = d_iob;

static struct d_token *loadhdr(),*parsehdr();
static int dumphdr();

void d_cleanup ()
{
  DATFILE **d_iop;
  
  for (d_iop = d_iob; d_iop < d_lastbuf; d_iop++)
    d_flush (*d_iop);
}

static void add_atexit ()
{
  /* Datlib uses d_cleanup() to ensure that all datfiles are closed when
     exit() is called.  add_atexit() is checked by d_open() and d_descr()
     to ensure that d_cleanup() is added to the atexit list once and only
     once. */

  static int added_atexit = 0;
  
  if (added_atexit == 0) {
#if defined (sun) && ! defined (SVR4)
    if (on_exit(d_cleanup) != 0) {
      fprintf (stderr, "libdat: unable to register d_cleanup() with on_exit()\n");
      fprintf (stderr, "libdat: datfiles will not be closed automatically on program exit.\n");
    }
#else
    if (atexit (d_cleanup) != 0) {
      fprintf (stderr, "libdat: unable to register d_cleanup() with atexit()\n");
      fprintf (stderr, "libdat: datfiles will not be closed automatically on program exit.\n");
    }
#endif
    added_atexit = 1;
  }
}

/* DATFILE *
 * d_open(name,mode)
 *    char *name,*mode; 
 * 
 *   is something like fopen().  It opens a datfile with access mode
 *   specified by the string mode.  (This may be "r","w","a","r+","w+","a+").
 *   Returns a pointer to struct of type DATFILE:
 *    DATFILE
 *    {
 *      struct d_token *db_h;		(ptr to token tree)
 *      FILE *db_fp;			(FILE pointer for data
 *      char *db_name;			(data file name specified in call)
 *      int db_flag;			(status flags such as _DB_WRHDR)
 *    };
 *   If mode="w":
 *	if name = old datfile name, or ordinary filename, deletes old file.
 *	if name = special file, creates simple output file, sets db_h = -1
 *	if name = other directory, then error.
 *	if name not exist, creates new dat file.
 *   If mode="r":
 *	if name not exist, then error.
 *	if name = old datfile, load header in descriptor file, open
 *		data file for buffered I/O, return struct of type DATFILE
 *	if name = old filename, open file for buffered I/O, return struct
 *		of type DATFILE with db_h = -1
 *   If mode="a":
 *	if name not exist, create new dat file.
 *	if name old datfile, load header, open data file for append
 *	if name old filename, open file for append, set db_h = -1
 *   If error occurs, return NULL.
 */

DATFILE *
d_open(prename,mode)
     const char *prename;	/* filename for picpath or full path */
     const char *mode;		/* mode for open */
{
  char buf[120];
  struct stat statbuf;		/* stats of file to open */
  int exist;			/* 0 if files exists, -1 if it doesn't */
  int flag;			/* 1 if file exists and is a directory, */
				/* 0 if file doesn't exist, and */
				/* -1 if file exists and isn't a directory */
  int wflag;			/* child's exit status */
  int type;
  char *D,*dname, name[256];
  DATFILE *d;
  int namlen,i;
  enum byte_order   bo ;

  add_atexit();

  /* Get the full path name from the picpath directory list. */
  d_path(prename,name);

  /* Check if filename exists, and its status */
  exist = stat(name,&statbuf);

  /* If we're opening for reading check for existence and permisissions */
  if (mode[0]=='r' || (mode[0]=='a' && exist != -1)) {

    /* Return an error if the file doesn't exist */
    if (exist == -1) {
      fprintf(stderr,"d_open: File %s does not exist\n",name);
      return ((DATFILE*)NULL);
    }

    /* If the file is a directory, it should be a datfile */
    if ((statbuf.st_mode & S_IFMT) == S_IFDIR) {

      /* make the descriptor filename */
      flag = 1;
      sprintf (buf, "%s%s", name, DESCRIPTOR);

      /* return error if descriptor doesn't exist */
      if (access (buf, 0)) {
	fprintf(stderr, "d_open: Directory %s not datfile\n",name);
	return ((DATFILE *) NULL);
      }
    }

    /* Set flag if this is just an ordinary file */
    else {
      flag = -1;
    }
  }

  /* If we're opening for writing, then blast old datfiles, but not other
     directories.  For appends, make sure this is a datfile */
  else if (mode[0]=='w' || mode[0]=='a') {

    /* Default action is to create a new directory */
    flag = 0;

    /* But if the file already exists... */
    if (exist != -1) {

      /* If it's a directory ... */
      if ((statbuf.st_mode & S_IFMT) == S_IFDIR) {
	sprintf(buf,"%s%s",name,DESCRIPTOR);

	/* Check for existence of a descriptor and erase old an datfile with
	   rm */
	if (! access (buf, 0)) {
	  if (vfork ())		/* parent waits for child to finish */
	    wait (&wflag);
	  else			/* child tries to delete directory */
	    execl("/bin/rm", "rm", "-r", name, NULL);

	  /* Return an error if the child didn't run successfully */
	  if (wflag) {
	    fprintf(stderr, "d_open: cannot delete datfile %s\n",name);
	    return((DATFILE*)NULL);
	  }
	}

	/* This is a directory, but not a datfile; return an error */
	else {
	  fprintf(stderr,"d_open: non datfile directory %s\n", name);
	  return ((DATFILE *) NULL);
	}
      }

      /* If it's a regular file, try to delete it with rm */
      else if ((statbuf.st_mode & S_IFMT) == (unsigned) S_IFREG) {
	if (vfork ())		/* parent waits for child */
	  wait (&wflag);
	else			/* child tries to delete file */
	  execl("/bin/rm", "rm", name, 0);

	/* Return an error if the child didn't return success. */
	if (wflag) {
	  fprintf(stderr, "d_open: cannot delete file %s\n",name);
	  return ((DATFILE*) NULL);
	}
      }

      /* If it's a special file, set flag. */
      else {
	flag = -1;
      }
    }

    /* The file doesn't exist, or flag says to create new directory. */
    if (exist == -1 || flag == 0) {

/*      if (vfork ())
	wait(&wflag);
      else
	execl("/bin/mkdir", "mkdir", name, 0);

      if (wflag) {
	fprintf(stderr, "d_open: Cannot make data directory %s\n",name);
	return ((DATFILE*) NULL);
      }
*/

      /* Try to create directory; return an error if unsuccessful. */
      if (mkdir (name, DEFAULT_DIR_PERMS)) {
	fprintf (stderr, "d_open: Cannot create directory %s.\n", name);
	return ((DATFILE *) NULL);
      }

      /* All went well; create the descriptor file. */
      sprintf (buf, "%s%s", name, DESCRIPTOR);
      i = creat (buf, DEFAULT_FILE_PERMS);
      if (i < 0) {
	/* error creating descriptor */
	perror("d_open") ;
	return ((DATFILE*) NULL);
      }

      close(i) ;
      flag = 1;			/* directory */
    }
  }

  /* File exists, but we can't overwrite it, so return an error. */
  else {
    fprintf(stderr, "d_open: unknown mode %s for file %s\n", mode, name);
    return ((DATFILE*) NULL);
  }

  /* allocate the DATFILE struct */
  d = (DATFILE *) calloc (1, sizeof (DATFILE));
  if (d == NULL) {
    perror("d_open");
    return ((DATFILE *) NULL);
  }

  /* If not dat file format, <name> is a data file.  Otherwise, the data
     file is <name>/DATA */
  if (flag == -1) D = "";
  else D = DATA;

  /* Allocate space for dname, and copy name of data file. */
  namlen = strlen (name);
  dname = malloc ((unsigned) (namlen + strlen(D) + 1) * sizeof(char));

  /* Check for an error in malloc. */
  if (dname == NULL) {
    perror("d_open");
    return ((DATFILE *) NULL);
  }

  strcpy(dname,name);
  strcat(dname,D);

  /* Create or open the data file. */
  if ((d->db_fp = fopen(dname, mode)) == NULL) {
    fprintf(stderr,"d_open: failed to fopen %s, mode %s\n",dname,mode);
    return ((DATFILE *) NULL);
  }

  /* Some machines don't let you write into the pre-exiting portion
     of a file which was opened "a+".  So, if the mode is "a+",
     then close the file at this point, re-open it "r+" and seek to
     the end of the file. */

  if (strcmp(mode, "a+") == 0)
    {
      fclose(d->db_fp);
      if ((d->db_fp = fopen(dname, "r+")) == NULL) {
	fprintf(stderr,"d_open: failed to fopen %s, mode %s\n",dname,mode);
	return ((DATFILE *) NULL);
      }
      fseek(d->db_fp, 0, SEEK_END);
    }

  /* Do the housekeeping on the struct */
  dname = realloc (dname, (unsigned) (namlen + 1) * sizeof(char)) ;
  dname[namlen] = '\0';
  d->db_name = dname;
  d->db_flag = 0;

  /* If a descriptor is available, generate a linked list of header
     entries and place its pointer in the db_h field.  Otherwise, just
     indicate that no header info is available */

  if (flag == -1)
    d->db_h = (struct d_token *) -1;
  else {
    d->db_h = loadhdr (name, mode);

    /* It ought to be an error if DESCRIPTOR file exists, but can't be
       loaded. */
    if (d->db_h == (struct d_token *) -1)
      return ((DATFILE *) NULL);
  }

  /* Bytes per sample defaults to zero. */
  d->db_nbs = 0;

  /* Put on tail of queue. */
  *d_lastbuf++ = d;

  /* Flush the buffers and get ready to check arch flags. */
  d_flush(d) ;

  /* Get the type of the datfile. */
  type = d_gettype(d, NULL) ;

  /* What checks are made depends on the file type.  
     If architecture (byte order) flag is non-existent, and
     we aren't creating a new dat file, warn the user.  If creating a new
     dat file, set the correct architecture.  If dat file flags conflicts
     with current machine, warn the user.  If dat file flags and machine
     agree, do nothing, we're okay.  */

  switch (type) {
  case 'f':			/* float  */ 
  case 'F':			/* double */
  case 'z':			/* complex float  */
  case 'Z':			/* complex double */
  case 'd':			/* 2 byte signed integer */
  case 'D':			/* 4 byte signed integer */
  case 'u':			/* 2 byte unsigned integer */
  case 'U':			/* 4 byte unsigned integer */
  case 'x':			/* 2 byte fixed point */
  case 'X':			/* 4 byte fixed point */
  case NOKEY:			/* unsure of type, check bo */
    bo = d_get_byte_order(d) ;

    /* Byte order flag non-existent... */
    if (bo == NOKEY) {

      /* If reading a file, give warning */
      if (mode[0] == 'r')
	fprintf(stderr, "d_open: Warning--no byte order flag.\n");

      /* If writing set endian flag */
      if (mode[0] == 'w')
	d_set_byte_order(d, mach_get_byte_order());

      /* If appending, ambiguity over opening or appending old file without
	 endian flags */
      if (mode[0] == 'a')
	fprintf(stderr,  "d_open: Warning--no byte order flag for append.\n");
    }

    /* byte order flag exists */
    else
      if (mach_get_byte_order() != bo) {
	fprintf(stderr,"ATTN: byte order conversion will occur.\n");
	/* exit (-1) ; */
      }
    break ;

  /* for the following cases, don't need to do any architecture checks */
  case 'a':			/* 1 byte ascii */
  case 'v':			/* 1 byte variant */
  case 'c':			/* 1 byte signed integer */
  case 'b':			/* 1 byte unsigned integer */
  default:			/* a type no ones heard of */
    break ;
  }

  d_flush(d) ;

  /* Return the dat file */
  return d;
}


/*
 * DATFILE *
 * d_descr(name)
 *	char *name;
 *
 *    Like d_open, but only reads in the header from the data file, and does
 *    not open the data file.  Returns a struct of type DATFILE with a pointer
 *    to the parsed token list, a null file pointer, and the name of
 *    the datfile.  If "name" isn't a directory, or if descriptor file can't
 *    be read, returns NULL.
 *
 */

DATFILE *
d_descr (name)
     const char *name;
{
  DATFILE* d;
	
  add_atexit();

  /* allocate the struct */
  d = (DATFILE *) calloc (1, sizeof (DATFILE));

  if (d == NULL) {
    perror("d_descr");
    return((DATFILE*)NULL);
  }

  d->db_h = loadhdr(name, "r");
  if (d->db_h == (struct d_token *) -1) {
    fprintf (stderr, "d_descr: header of %s not found\n", name);
    free ((char *) d);
    return ((DATFILE*) NULL);
  }

  d->db_name = malloc ((unsigned) strlen(name) + 1);
  if (d->db_name == NULL) {
    perror ("d_descr");
    free ((char *) d);
    return ((DATFILE *) NULL);
  }

  strcpy (d->db_name, name);
  d->db_fp = (FILE*) NULL;
  d->db_flag = 0;
  d->db_nbs = 0;
  *d_lastbuf++ = d;		/* put on tail of queue */
  return(d);
}	


/* int
 * d_flush(d)
 *	DATFILE *d;
 *
 *	the equivalent of fflush().  Dumps the header into the descriptor file
 *	if the header has been modified, and clears the modified bit.  Flushes
 *	the buffered output file.
 *  
 * MODIFIED: 3/23/89 by Pat Romano: added check for files not open for
 *			writing.
 */

int d_flush (d)
     DATFILE *d;
{
  int i = 0;
  int j = 0;
  FILE *fp;
    
  if (d->db_flag & _DB_WRHDR) {
    i = dumphdr(d);
    d->db_flag &= ~_DB_WRHDR;	/* clear modified header bit */
  }

  fp = d_getfp(d);

#ifndef linux
  if ((fp != NULL) && (fp != (FILE *) - 1) && 
      ((fp->_flag & _IOWRT) || (fp->_flag & _IORW) 
#ifdef _IOAPPEND
       || (fp->_flag & _IOAPPEND)
#endif
       ))
#else /* linux */
  if ((fp != NULL) && (fp != (FILE *) - 1))
#endif
    {
      if ((j = fflush (fp)) == EOF) {
	perror(d->db_name);
      }
    }

  return(i + j);
}



/* int
 * d_close(d)
 * 	DATFILE *d;
 * 
 * 	the equivalent of fclose().  Dumps the header into the descriptor
 *	file if the header was modified (db_flag bit _DB_WRHDR set), then
 *	fclose() the data file pointer, and frees all malloc() storage used in
 *	the DATFILE struct, the name, and the token list.
 */

int d_close (d)
     DATFILE *d;
{
  int i = 0;
  int j = 0;
  struct d_token *tp,*tp2;
  DATFILE **d_iop;

  /* search queue for datfile */	
	
  for (d_iop = d_iob; d_iop < d_lastbuf && *d_iop != d; d_iop++)
    ;

  if (d_iop >= d_lastbuf) {		/* non-existent datfile */
    fprintf (stderr,"d_close: File not open!?\n");
    return (-1);
  }
  else {
    *d_iop = *--d_lastbuf;		/* move tail of queue to vacancy */
  }

  if (d->db_flag & _DB_WRHDR)
    i = dumphdr (d);

  if (d->db_fp != NULL) {
    _dclrinfo (fileno (d_getfp (d)));	/* notify datio package of d_close */
    if ((j = fclose (d->db_fp)) != 0)
      perror(d->db_name);
  }

  if (d->db_h != (struct d_token *)-1) {
    for (tp = d->db_h; tp; ) {
      free ((char *) tp->d_key);
      free ((char *) tp->d_value);
      tp2 = tp;
      tp = tp->d_next;
      free ((char *) tp2);
    }
  }
  free ((char *) d->db_name);
  free((char * ) d);
  return(i + j);
}



/* loadhdr(name,mode)
 *
 * takes the string passed as the argument, which should be the name
 * of the datfile directory, and opens up the file DESCRIPTOR within
 * it, and parses the token list it finds there.  It returns a pointer
 * to a linked list of tokens.  If mode is "r" and can't find DESCRIPTOR,
 * returns -1.  If mode is "w" or "a" and can't find DESCRIPTOR, creates
 * an empty descriptor file and returns NULL.  If error, returns -1.
 * loadhdr() should only be called if we are really working on a dat file
 * structure.
 */

static struct d_token *
loadhdr(name,mode)
     char *name,*mode;
{
  char *hname;
  FILE *fp;
  struct d_token *ptr ;
  int i ;
  char* D = DESCRIPTOR;

  /* Open the descriptor */

  hname = malloc ((unsigned) (strlen(name) + strlen(D) + 1) * sizeof(char));
  if (hname == NULL) {
    perror("loadhdr");
    return((struct d_token*)-1);
  }

  strcpy (hname, name);
  strcat (hname, D);
  fp = fopen (hname, "r");
  if (fp == NULL && *mode=='r') {
    perror("loadhdr") ;	/* Without this line, the error would be silent. */
    ptr = (struct d_token*)-1 ;
  }
  else if (fp == NULL) {
    if ((i = creat(hname,0644)) == -1) {
      perror("loadhdr") ;
      ptr = (struct d_token*) -1 ;
    }
    else {
      close(i) ;
      ptr = (struct d_token*)NULL ;
    }
  }
  else {
    ptr = parsehdr(fp) ;
    fclose(fp) ;
  }
  free(hname);
  return(ptr) ;
}

/* parsehdr(fp)
 *	FILE *fp;
 *
 *	Actually parses the header, and returns a pointer to the token
 *	list.  It wants a file pointer opened to the header.  Calls malloc()
 *	to allocate each key and value string, and each token struct.
 */

static struct d_token *
parsehdr (fp)
     FILE *fp;
{
#define PARSEBUF 256
  char *buf;
  unsigned int buflen;
  int c;
  char *p,*q;
  struct d_token *tp;
  struct d_token dummy;
  int level;	/* for keeping track of nesting of TOKEN chars */
  int comment;

  /* Now parse the descriptor */
  dummy.d_next = (struct d_token*) NULL;
  tp = &dummy;
  level = comment = 0;
  c = getc (fp);
	
  while (c != EOF) {
    if (c == LTOKEN) {
      level++;
      if ((tp->d_next = (struct d_token *)
	   calloc(1, sizeof(struct d_token))) == NULL)
	{
	  perror("parsehdr");
	  return(dummy.d_next);
	}
      tp = tp->d_next;

      /* Get the key */
      while((c=getc(fp)) != EOF && isspace(c));	/* skip space */
      buflen = PARSEBUF;

      /* get buffer */
      if ((buf=q=malloc(buflen)) == (char*)NULL) {
	perror("parsehdr") ;
	return(dummy.d_next) ;
      }

      /* copy chars until EOF, space, RTOKEN */
      while (c!=EOF && !isspace(c) && c!=RTOKEN) {
	if (isupper(c))			/* convert to lower case */
	  c = tolower(c);
	if (buflen-- == 0 && addch(&buf,&q,&buflen) < 0)
	  return (dummy.d_next) ;
	*q++ = c ;
	c = getc(fp);
      }

      if (buflen-- == 0 && addch(&buf,&q,&buflen) < 0)
	return (dummy.d_next) ;
      *q++ = '\0';
      if ((tp->d_key=realloc(buf,(unsigned)(q-buf))) == NULL) {
	perror("parsehdr");
	return(dummy.d_next);
      }

      /* Collect value list */
      if(isspace(c))
	c = getc(fp);	/* skip over only one space */

      /* get buffer */
      buflen = PARSEBUF;
      if ((buf=q=malloc(buflen)) == (char*)NULL) {
	perror("parsehdr") ;
	return(dummy.d_next) ;
      }

      while (c != EOF && level) {
	switch (c) {
	case LTOKEN:
	  level++;
	  if(buflen-- == 0 && addch(&buf,&q,&buflen) < 0)
	    return(dummy.d_next) ;
	  *q++ = c ;
	  break;
	case RTOKEN:
	  if (--level>0) {
	    if (buflen-- == 0 && addch(&buf,&q,&buflen) < 0)
	      return(dummy.d_next) ;
	    *q++ = c ;
	  }	/* skip over terminating RTOKEN (level==0)*/
	  break;
	case ESCCHAR:
	  if ((c = getc(fp)) != EOF) {
	    if (buflen-- == 0 && 
	       addch (&buf, &q, &buflen) < 0)
	      return (dummy.d_next) ;
	    *q++ = c;
	  }
	  break;
	default:
	  if (buflen-- == 0 && addch(&buf,&q,&buflen) < 0)
	    return (dummy.d_next) ;
	  *q++ = c;
	  break;
	}
	c = getc(fp);
      }

      if (buflen-- == 0 && addch(&buf,&q,&buflen) < 0)
	return(dummy.d_next) ;
      *q++ = '\0' ;
      if ((tp->d_value=realloc(buf,(unsigned)(q-buf))) == NULL) {
	perror("parsehdr");
	return(dummy.d_next);
      }

      if (level != 0)
	goto balance;
    }

    else while (c != LTOKEN && c!=EOF) {
      switch (c) {
      case RTOKEN:
	if (!comment)
	  goto balance;
	break;
      case COMCHAR:
	comment = 1;
	break;
      case '\n':
	comment = 0;
	break;
      default:
	if (!isspace(c) && !comment) {
	  fprintf(stderr, "parsehdr: Illegal character %c outside keys\n", c);
	  return (dummy.d_next);
	}
	break;
      }
      c = getc(fp);
    }
  }
	
  return(dummy.d_next);

 balance:
  fputs("parsehdr: Unbalanced LTOKEN and RTOKENs\n",stderr);
  return(dummy.d_next);
}


/* Expands parsehdr() buffer by PARSEBUF chars
   Returns -1 if malloc error, 0 if OK  */

int addch(buf,q,buflen)
     char **buf,**q ;
     unsigned int *buflen ;
{
  char *ptr ;
  unsigned int len = *q - *buf ;
  *buflen += PARSEBUF ;

  if ((ptr=realloc(*buf,(unsigned)(len + PARSEBUF))) == (char*)NULL)
    {
      perror("parsehdr") ;
      return(-1) ;
    }
  *buf = ptr ;
  *q = ptr+len ;
  return(0) ;
}


/* dumphdr(d)
 *
 *	writes the token list in d->db_h to the file d->db_name/DESCRIPTOR.
 *	Returns -1 if error.
 */

static int 
dumphdr (d)
     DATFILE *d;
{
  struct d_token *tp;
  char *hname;
  FILE *hf;
  char *D = DESCRIPTOR;
  int flag;

  /* If an ordinary file, can't dump header */
  if((tp=d->db_h) == (struct d_token*) -1)
    return(NULL);
	
  /* Open up the descriptor */
  if ((hname = malloc((unsigned)(strlen(d->db_name)+strlen(D)+1)
		      *sizeof(char))) == NULL)
    {
      perror("dumphdr");
      return(-1);
    }

  strcpy(hname,d->db_name);
  strcat(hname,D);
  hf = fopen(hname,"w");
  if (hf == NULL) {
    perror(hname);
    free(hname);
    return(-1);
  }

  for (; tp!=NULL ; tp=tp->d_next) {
    if (tp->d_key) {
      register char *s;

      putc(LTOKEN,hf);
      fputs(tp->d_key,hf);
      if (tp->d_value) {
	putc(' ',hf);
	for (s=tp->d_value; *s ; s++) {
	  switch(*s) {
	  case LTOKEN:
	  case RTOKEN:
	  case ESCCHAR:
	  case COMCHAR:
	    putc(ESCCHAR,hf);
	  default:
	    putc(*s,hf);
	  }
	}
      }
      putc(RTOKEN,hf);
      putc('\n',hf);
    }
  }
  flag = 0;

  if (fclose(hf) == EOF) {
    perror(hname);
    flag = -1;
  }
  free(hname);
  return(flag);
}
