/*
  d_path()
  
  d_path will take a "picture" name-- that is, a string like "title" or
  "/u/pix/pam" and return an integer representing the existance and/or
  format of the data.
  
  d_open will, given a simple picture title without a specified path,
  search through possible paths for that picture, or else it will search
  for the picture through the path specified, if any.
  
  If a path is specified,
  * the integer 1 is returned if the picture
  exists there,

  If no path is specified,
  * the integer 1 is returned if path is found
  to be: "/u/pix/title/data", e.g. a DAT file

  * 0 (zero) is returned if it does not exist in either case. 

  picopen looks at the environment variable PICPATH to determine where
  to open the picture file.  You should have something along the lines
  of `setenv PICPATH ( /u/pix . /u/cm )' in your .cshrc file on the
  SUNs, and prehaps `setenv PICPATH ( /suns/u/pix /pix/suns . )' on the
  VAX.
  
  When it cannot open a file (for whatever reason), picopen returns a
  negative number.
  
  Examples:
  
  */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <sys/errno.h>
#include <sys/fcntl.h>

#define DEFAULTPATH ".:"

static  char *defaultpath = DEFAULTPATH;
extern int	errno;

/*
 * look for a '/' in a string.  if it's there, assume we are dealing with a
 * path.
 */
int d_ispath( s )
char *s;
{
  while( *s != '\0' ) if( *s++ == '/' ) return( 1 );
     return( 0 );
}

/*
 *
 */
int d_path(picname,picturepath)
char *picname,*picturepath;
{ 
     int flags = O_RDONLY;
     struct stat	statbuf;
     char dirname[256];
     char pic_path[256], *ppath, *getenv();

     /* are we dealing with a path or a basename? */
     if( d_ispath( picname ) == 0 ) 
       {
	    /* If picname is basename,
	       walk through PICPATH til you find picture */
	    if( (ppath = getenv( "PICPATH" )) == NULL )
	      /* nothing, then use `.' */
	      ppath = defaultpath;
	    /* read from ppath upto colon and put results into picpath */
	    while( sscanf( ppath, "%[^:]", pic_path ) == 1 )
	      { /* walk PICPATH */
		   /* first, try to open picpath/picname/data */
		   strcpy( dirname, pic_path );
		   strcat( dirname, "/" );
		   strcat( dirname, picname );
		   strcat( dirname, "/" );
		   strcat( dirname, "data" );
		   if( (stat( dirname, &statbuf ) == 0) &&
		      ((statbuf.st_mode & S_IFMT) != S_IFDIR) ) 
		     {
		       strncpy( picturepath, dirname,
			       strlen(dirname)-strlen("/data"));
		       picturepath[strlen(dirname)-strlen("/data")]=0;
		       return( 1 );
		     }
		   /* Drat! No luck! Try next picpath */
		   ppath += (strlen( pic_path ));
		   /* if null, quit, else skip over the colon */
		   if( *ppath != '\0' ) ppath++;

	      }  /* end while. tried all paths in PICPATH */

	    /* if get to this point, picture was not found. */

	    /* return 0 and make /pix/sun/ the path for the new pic. */
	    if ( (ppath = getenv("PICPATH"))==NULL) 
	      ppath = defaultpath;
	    if (sscanf(ppath, "%[^:]", pic_path) != 1)
		return (-1);
		strcpy (picturepath, pic_path);
		strcat (picturepath, "/");
		strcat (picturepath, picname);
                return (0);

       }      /* end if picname is basename */
     /******************************/
   
     else  /* it is a path */ 
       {
	    strcpy (dirname,picname);
	    strcat (dirname,"/");
	    strcat (dirname, "data");
	    /* make sure we've not opened a directory */
	    if ( (stat(dirname, &statbuf ) == 0) &&
		((statbuf.st_mode & S_IFMT) != S_IFDIR) )
	      {
		strncpy( picturepath, dirname,
			strlen(dirname)-strlen("/data"));
		picturepath[strlen(dirname)-strlen("/data")]=0;
		return( 1 );
	      }
	  }
	      
     /* if get to here, the pic doesn't exist in given path. */

     /* just return the given path as picturepath */
	
	strcpy( picturepath, picname);
        return( 0 );
   }

/* d_path */
