//
// Sound.cc
// 
// C++ interface to soundfiles
// based on snd.c
// dpwe@media.mit.edu 1994nov02
//

#include "Sound.h"
#include <assert.h>
#include <stdio.h>

Sound::Sound(char modechr, double sampRate, int chans, Format fmt)
{   /* Create a new sound object.  Default arguments specified for 
       each parameter in header file (the C++ way). */
    if(modechr != 'r' && modechr != 'w')  {
	// Called illegally
	delete this;
    } else {
	snd = sndNew(NULL);
	snd->snd->samplingRate = sampRate;
	snd->snd->channels   = chans;
	snd->snd->format     = fmt;
	snd->mode = modechr;
    }
}

Sound::~Sound(void)
{   /* close the soundfile, destroy the sound object */
    if(snd) {
	sndFree(snd);
    }
}

int SfmtSize(int format)
{   /* return the bytes per sample of a particular data format */
    return (format & 0x0F);	// for now, in lower 4 bits of format word
}

const int BITS_PER_BYTE = 8;

int Sound::open(char *filename)
{   /* open file for input, or create for output.  Returns zero on success. */
    char modeStr[2];

    modeStr[0] = snd->mode;	modeStr[1] = '\0';
    if( (snd=sndOpen(snd,filename,modeStr))==NULL)  {
	return -1;
    } else {
	return 0;
    }
}

long Sound::readFrames(void *buf, long frames)
{   /* read sound data from file */
    return sndReadFrames(snd, buf, frames);
}

long Sound::writeFrames(void *buf, long frames)
{   /* write sound data to file */
    return sndWriteFrames(snd, buf, frames);
}

long Sound::frameTell(void)
{   /* frame index of next to read */
    return sndFrameTell(snd);
}

long Sound::frameSeek(long frame)
{   /* move file point to abs pos (in sample frames) */
    return sndFrameSeek(snd, frame, SEEK_SET);
}

int Sound::frameFeof(void)
{   /* return nonzero if at EOF */
    return sndFeof(snd);
}

