/***************************************************************\
*   sndf.c		    					*
*   implementation of for sound file interface			*
*   just writes files as images of memory block			*
*   'inspired' by the NeXT SND system				*
*   10aug90 dpwe						*
\***************************************************************/

/* #define DEBUG 1		/* diagnostics during read */

/* #define PCM		1	/* read headerless Unix files */
#define MACAIFF	1	/* mac writes as AIFF files (else sdII) */
#define AUDIOMEDIA 1	/* SoundDesigner files written for audiomedia board */

#include <stdio.h>	    /* ANSI file ops */
#include <string.h>	    /* for buildsfname only */

#ifdef __sgi
/* #define AIFF */
/* #undef AIFF
   #define NeXT
 */
#endif /* __sgi */		/* AIFFs on indigo */

#ifdef sun
#define  strchr(A,B) 	index(A,B)
#define  strrchr(A,B)	rindex(A,B)
#define  NeXT	/* use NeXT soundfiles on sun */
#endif

#ifdef THINK_C
/* #define __STDC__ */ /* predefined in THINKC5 */
#include <stdlib.h>	    /* malloc stuff */
#define READMODE "rb"
#define WRITEMODE "wb+"     /* "+" allows infoBsize reread */
#else
#ifdef NeXT
#include <stdlib.h>
#else
#include <sys/types.h>
#include <unistd.h>	    /* SEEK_SET etc */
#include <malloc.h>	    /* malloc stuff */
#endif
#define READMODE "rb"	    /* was + for read-mod-write, but that's a lose */
#define WRITEMODE "wb+"	    /* "+" allows infoBsize reread */
#endif

/* Prototype argument wrapper */
/* make fn protos like   void fn PARG((int arg1, char arg2));  */
#ifdef __STDC__
#define PARG(a)		a
#else /* !__STDC__ */
#define PARG(a)		()
#endif /* __STDC__ */

#include <sndf.h>
#define READ_FLAG	1
#define WRITE_FLAG	0	/* flags to FixupSamples (byte ordering) */

/* static variables */

static SFSTRUCT tmphdr;     /* scratch space for pre-load */
/* static */ int	SFerror;    /* (internal) error flag */
static char	unspecMsg[] = "Unspecified error : _______";
/* want to 'fill in' at    012345678901234567890... [20..27] */
#define USMINDX 20	/* from above, chr index into string */

#ifdef __STDC__
/* static function prototypes */
static void itoa(int n, char *s);   /* for unrecognised error code */

/* more static function prototypes - these fn's in personality #incl */
static void AddFname(char *fname, FILE *file);	/* filename table */
static void RmvFname(FILE *file);	/* implemented if needed */
static long SFHeDaLen(FILE *file);  /* return space used for fi hdr */
static void FixupSamples(void *ptr, int format, long nitems, int flag);
                        /* hook for byte reordering */
#else /* !__STDC__ */
static void itoa();
static void AddFname();
static void RmvFname();
static long SFHeDaLen();
static void FixupSamples();
#endif /* !__STDC__ */

/* ----------------------- implementations ------------------------ */

#ifdef THINK_C
#ifdef MACAIFF
#include "sndfaif.c"		/* AIFF version */
#else /* not MACAIFF -> SoundDesigner */
#include "sndfsd2.c"	    /* mac versions of header r/w &c */
#endif /* MACAIFF */
#else  /* !THINK_C */
/* #include "sndfnat.c"	    /* native (direct memory image) headers */
#ifdef NeXT
#include "sndfnxt.c"	    /* NeXT '.snd' soundfiles */
#else /* !NeXT */
#ifdef PCM		    /* headerless files */
#include "sndfpcm.c"
#else /* !PCM */
#ifdef AIFF
#include "sndfaif.c"	    /* AIFF on UNIX for Indigo */
#else /* !AIFF */
#ifdef MS_WAVE
#include "sndfwav.c"	    /* Microsoft WAV/RIFF PC format */
#else /* !MS_WAVE */
#include "sndfirc.c"	    /* standard IRCAM headers */
#endif /* !MS_WAVE */
#endif /* !AIFF	*/
#endif /* !PCM	*/
#endif /* !NeXT	*/
#endif /* !THINK_C	*/

int SFInfoSize(snd)
    SFSTRUCT *snd;	/* return number of info bytes in header */
    {
    return(snd->headBsize - sizeof(SFSTRUCT) + SFDFLTBYTS);
    }

char *SFDataLoc(snd)
    SFSTRUCT *snd;	/* return pointer to data from header */
    {
    return( ((char *)snd + snd->headBsize) );
    }

/* return number of bytes per sample in given fmt */
/* 
int SFSampleSizeOf(format)
    int format;	
    {
    return(format & SF_BYTE_SIZE_MASK);
    }
 */
/* now in sndfutil.c */

int SFSampleFrameSize(snd)
    SFSTRUCT *snd;	/* return number of bytes per sample frame in snd */
    {
    return snd->channels * SFSampleSizeOf(snd->format);
    }

int  SFAlloc(psnd, dataBsize, format, srate, chans, infoBsize)
    SFSTRUCT **psnd;
    long dataBsize;
    int format;
    FLOATARG srate;
    int chans;
    int infoBsize;
/*  Allocate memory for a new SFSTRUCT+data block;fill in header
    Returns SFE_MALLOC (& **psnd = NULL) if malloc fails    */
    {
    long    bSize;
    long    infoXtra;

    if(infoBsize < SFDFLTBYTS)	infoBsize = SFDFLTBYTS;
    infoBsize = (infoBsize + 3)&-4; /* round up to 4 byte multiple */
    infoXtra = (long)(infoBsize-SFDFLTBYTS);
    if(dataBsize == SF_UNK_LEN)
	bSize = sizeof(SFSTRUCT) + infoXtra;
    else
	bSize = dataBsize + sizeof(SFSTRUCT) + infoXtra;
    if( ( (*psnd) = (SFSTRUCT *)malloc((size_t)bSize)) == NULL)
	return(SFerror = SFE_MALLOC);
    (*psnd)->magic = SFMAGIC;
    (*psnd)->headBsize = sizeof(SFSTRUCT) + infoXtra;
    (*psnd)->dataBsize = dataBsize;
    (*psnd)->format    = format;
    (*psnd)->samplingRate = srate;
    (*psnd)->channels  = chans;
    /* leave info bytes undefined */
    return(SFerror = SFE_OK);
    }

void SFFree(snd)
    SFSTRUCT *snd;	/* destroy in-memory sound */
    {
    free(snd);
    }

FILE *SFOpenRdHdr(path, snd, infoBmax)
    char *path;
    SFSTRUCT *snd;
    int infoBmax;
/*  Opens named file as a stream, reads & checks header, returns stream
    Attempts to read up to infoBmax info bytes, at least SFDFLTBYTS.
    No error codes; returns NULL on fail (SFE_NOPEN, _NSF, _RDERR) */
    {
    FILE    *file;

#ifdef DEBUG
	fprintf(stderr,"SFOpRdH: entered for %s\n", path);
#endif
    if(strcmp(path, SF_STDIN_STR)==0)  {   /* special case */
        file = stdin;
    } else if( (file = fopen(path, READMODE)) == NULL)
	{
#ifdef DEBUG
	fprintf(stderr,"SFOpRdH: fopen (mode %s) failed\n", READMODE);
#endif
	SFerror = SFE_NOPEN;
	return(NULL);
	}
    AddFname(path,file);	/* create filename table entry */
    SFerror = SFReadHdr(file, snd, infoBmax);
#ifdef DEBUG
	fprintf(stderr,"SFOpRdH: RdH returned %d\n", SFerror);
#endif
    if(SFerror == SFE_OK)
	return(file);
    else
	{
	fclose(file);
	RmvFname(file);		/* remove filename table entry */
	return(NULL);
	}
    }

FILE *SFOpenAllocRdHdr(path, psnd)
    char *path;
    SFSTRUCT **psnd;
     /* find sizeof header, alloc space for it, read it in.
        returns NULL on fail (SFE_NOPEN, _NSF, _RDERR) */
    {
    FILE	*retFP;
    int 	i;
    long	headBsz;

    if(strcmp(path, SF_STDIN_STR)==0)  {   /* special case */
        retFP = stdin;
    } else if( (retFP = fopen(path, READMODE)) == NULL)
	{
	SFerror = SFE_NOPEN;
	return NULL;
	}
    AddFname(path,retFP);	/* create filename table entry */
    if( (SFerror = SFAllocRdHdr(retFP, psnd)) != SFE_OK)
	{
	fclose(retFP);
	RmvFname(retFP);
	return NULL;
	}
    else
	return retFP;
    }

FILE *oldSFOpenAllocRdHdr(char *path, SFSTRUCT **psnd);	/* ob.proto for THINK_C */

FILE *oldSFOpenAllocRdHdr(path, psnd)
    char *path;
    SFSTRUCT **psnd;
     /* find sizeof header, alloc space for it, read it in.
        returns NULL on fail (SFE_NOPEN, _NSF, _RDERR) */
    {
    FILE	*retFP;
    int 	i;
    long	headBsz;

    retFP = SFOpenRdHdr(path, &tmphdr, SFDFLTBYTS);
    if(retFP != NULL)
	{
	headBsz = tmphdr.headBsize;
	*psnd = (SFSTRUCT *)malloc(headBsz);
	if(*psnd != NULL)	/* ..else ukky malloc fail */
	    {
	    for(i = 0; i<sizeof(SFSTRUCT); ++i)	/* bytewise copy */
		((char *)*psnd)[i] = ((char *)&tmphdr)[i];
	    SFRdXInfo(retFP, *psnd, SFDFLTBYTS); 
	    }
	}
    return retFP;
    }

int SFAllocRdHdr(file, psnd)	/* allocate & read a header from open file */
    FILE *file;
    SFSTRUCT **psnd;
    {
    long headBsz;
    int  i;

    if( file == NULL)
	return(SFerror = SFE_NOPEN);
    if( (SFerror = SFReadHdr(file, &tmphdr, SFDFLTBYTS)) != SFE_OK)
	return(SFerror);
    headBsz = tmphdr.headBsize;
    *psnd = (SFSTRUCT *)malloc(headBsz);
    if(*psnd != NULL)	/* ..else ukky malloc fail */
	{
	for(i = 0; i<sizeof(SFSTRUCT); ++i)	/* bytewise copy */
	    ((char *)*psnd)[i] = ((char *)&tmphdr)[i];
	SFRdXInfo(file, *psnd, SFDFLTBYTS); 
	}
    else
	return(SFerror = SFE_MALLOC);

    return(SFerror = SFE_OK);
    }

FILE *SFOpenWrHdr(path, snd)
    char *path;
    SFSTRUCT *snd;
/*  Creates file in path & writes out header, returns stream for access
    No error codes, just returns NULL on fail (SFE_NSF, _NOPEN, _WRERR)  */
    {
    FILE    *file;

    if(snd->magic != SFMAGIC)
	{
	SFerror = SFE_NSF;
	return(NULL);
	}
    if(strcmp(path, SF_STDOUT_STR)==0)  {   /* special case */
        file = stdout;
    } else if( (file = fopen(path, WRITEMODE)) == NULL)
	{
	SFerror = SFE_NOPEN;
	return(NULL);	
	}
    AddFname(path,file);	/* create a filename table entry */
    if(SFWriteHdr(file,snd,0))
	{
	fclose(file);
	RmvFname(file);		/* remove filename table entry */
	SFerror = SFE_WRERR;
	return(NULL);	
	}
    else
	{
	SFerror = SFE_OK;
	return(file);
	}
    }

void SFCloseWrHdr(file,snd)
    FILE *file;
    SFSTRUCT *snd;
/*  After opening via SFOpenWrHdr and writing sound to the stream, may
    need to update header with a postieri data (sound length, more info).
    SFCloseWrHdr works 3 ways:
	- snd is NULL	    look at file ptr, rewrite length to hdr
	- snd exists, but dataBlen is 0
			    fill in new dataBlen from file ptr, write hdr
	- snd exists including dataBlen
			    rewrite whole header
    Ignores any fails & just closes file as best it can.
    Must have readback access for file (mode "w+"), else can 
    trash the header (e.g. if opened "w" under ultrix 4.0) */	
    {
    long len;

    len = ftell(file);
    if(len >= 0)		/* ONLY if can seek i.e. not pipe */
	{
	len -= SFHeDaLen(file);    /* knock off written hdr */
	if(snd == NULL)
	    {
	    SFerror = SFReadHdr(file, &tmphdr, SFDFLTBYTS); /* read orig hdr */
	    tmphdr.dataBsize = len; /* fixup data length */
	    /* rewrite modified header, but only as many infobyts as we read */
	    if(SFerror == SFE_OK) /* only write back if we read good header */
		SFerror = SFWriteHdr(file, &tmphdr, SFDFLTBYTS);
	    }
	else{	/* we were passed a header to rewrite */
	    if(snd->dataBsize == 0 || snd->dataBsize == SF_UNK_LEN)
		/* .. but does it have a length ? */
		snd->dataBsize = len; /* fill in if not */
	    SFerror = SFWriteHdr(file, snd, 0); 
	    /* write out whole lot (incl. all info) */
	    }
	}
    fclose(file);
    RmvFname(file);	/* remove filename table entry */
    }

int SFRead(path,psnd)
    char *path;
    SFSTRUCT **psnd;
/* Opens & checks file specified in path.  Allocs for & reads whole sound.
   Poss errors: SFE_NOPEN, SFE_NSF, SFE_MALLOC, SFE_RDERR */
    {
    FILE    *file;
    int     i, err = SFE_OK;
    int     xinfo;
    int     format = 0;
    int     chans  = 0;
    float   srate  = 0.0;
    size_t  num;

#ifdef DEBUG
	fprintf(stderr,"SFRd: reading %s\n", path);
#endif
    if( (file = SFOpenRdHdr(path, &tmphdr, SFDFLTBYTS)) == NULL)
	return(SFerror);
#ifdef DEBUG
	fprintf(stderr,"SFRd: hdr of %s read\n", path);
#endif
    xinfo = tmphdr.headBsize - sizeof(SFSTRUCT);
    err = SFAlloc(psnd, tmphdr.dataBsize, format, srate, chans,
		    xinfo + SFDFLTBYTS);
#ifdef DEBUG
	fprintf(stderr,"SFRd: alloc rtns %d\n", err);
#endif
    if(err == SFE_OK)
	{
	long ssiz;

	/* copy over what we already read */
	for(i = 0; i < sizeof(SFSTRUCT)/2; ++i)
	    ((short *)*psnd)[i] = ((short *)&tmphdr)[i];
	/* grab remaining info chars (may be 'elsewhere') */
	if( xinfo > 0 )
	    SFRdXInfo(file, *psnd, SFDFLTBYTS);
	format = (*psnd)->format;
	ssiz = SFSampleSizeOf(format);
	/* read all the data */
	if((*psnd)->dataBsize != SF_UNK_LEN)	/* i.e. a valid length */
	    if( (num = SFfread(SFDataLoc(*psnd),
			     format, 
			     (size_t)(*psnd)->dataBsize/ssiz,
			     file))
	       < (*psnd)->dataBsize/ssiz )
		{
		fprintf(stderr,"SFRead: wanted %ld got %ld of %ld\n",
			(*psnd)->dataBsize/ssiz, num, ssiz);	
		err = SFE_RDERR;
		}
	}
    if( (err != SFE_OK) && (*psnd != NULL))
	{
	SFFree(*psnd);
	*psnd = NULL;
	}
    fclose(file);
    RmvFname(file);	/* remove filename table entry */
    return(SFerror = err);
    }

int SFWrite(path,snd)
    char *path;
    SFSTRUCT *snd;
/* Write out the soundfile to the file named in path.
   Poss errors: SFE_NSF, SFE_NOPEN, SFE_WRERR */
    {
    FILE    *file;
    int     err = SFE_OK;
    long    bSize, items;

    if(snd->magic != SFMAGIC)
	return(SFerror = SFE_NSF);
    if( (file = SFOpenWrHdr(path, snd)) == NULL)
	return(SFerror);
    bSize = snd->dataBsize;
    if(bSize != SF_UNK_LEN)  {	/* i.e. a valid length */
	items = bSize / SFSampleSizeOf(snd->format);
	if( SFfwrite(SFDataLoc(snd), snd->format, items, file)
	   < items )
	    err = SFE_WRERR;
    }
    fclose(file);
    RmvFname(file);	/* remove filename table entry */
    return(SFerror = err);
    }

char *SFErrMsg(err)
    int err;     /* convert error code to message */
    {
    switch(err)
	{
    case SFE_OK:	return("No SF error");
    case SFE_NOPEN:	return("Cannot open sound file");
    case SFE_NSF:	return("Object/file not sound");
    case SFE_MALLOC:	return("No memory for sound");
    case SFE_RDERR:	return("Error reading sound file");
    case SFE_WRERR:	return("Error writing sound file");
    default:
	itoa(err, &unspecMsg[USMINDX]);
	return(unspecMsg);
	}
    }

void SFDie(err, msg)	/* what else to do with your error code */
    int  err;
    char *msg;
    {
    if(msg != NULL && *msg)
	fprintf(stderr,"%s: error: %s (%s)\n",programName,SFErrMsg(err),msg);
    else
	fprintf(stderr,"%s: error: %s\n",programName,SFErrMsg(err));
    exit(1);
    }

/* itoa from K&R p59 */
static void itoa(n,s)
    int n;
    char *s;
    {
    int     i,j,sign;
    char    c;

    if( (sign = n) < 0)     n = -n;
    i = 0;
    do	{
	s[i++] = n%10 + '0';
	} while( (n /= 10) > 0);
    if(sign < 0)
	s[i++] = '-';
    s[i--] = '\0';
    for(j=0; j<i; ++j,--i)	/* reverse the string */
	{ c = s[i]; s[i] = s[j]; s[j] = c; }
    }

/* buildsfname stolen from csound */

#define MAXFULNAME 128

char *buildsfname(name)			/*  build a soundfile name	*/
    register char *name;       		/*  <mod of carl getsfname()>	*/
    {
    static	char	fullname[MAXFULNAME];
#ifdef THINK_C
/*    extern char sfdir_path[];	*/
    /*
     * If name is not a pathname, then construct using sfdir_path[].
     */
    fullname[0] = 0;
    if (strchr(name,':') == NULL) {
	strcpy(fullname,sfdir_path);
	strcat(fullname,":");
	}
    strcat(fullname,name);
    return fullname;
#else
    char	*sfdir, *getenv();

    if (*name == '/' || *name == '.') /* if path already given */
	return(strcpy(fullname,name)); /*	copy as is	*/
    else {			/* else build right here  */
	if ((sfdir = getenv("SFDIR")) != NULL)
	    {
	    sprintf(fullname,"%s/%s",sfdir,name);
	    return(fullname);
	    }
	else
	    {
	    /* die("buildsfname: environment variable SFDIR undefined"); */
	    strcpy(fullname,name);
	    return(fullname);
	    }
	}
#endif
    }

/* covers for other major ANSI file functions
   - allow filesize faking and invisible byte reordering (for AIFF) */

size_t SFfread(ptr, format, nitems, stream)
    void *ptr;
    int  format;
    size_t nitems;
    FILE *stream;
    {
    size_t	size;
    long	ret;
    long 	pos, end, remBytes;

    size = SFSampleSizeOf(format);
    pos = ftell(stream);
    if(pos >= 0)
	{
	end = SFLastSampTell(stream);
	if(end > 0)
	    {
	    remBytes = end - pos;
	    /* asking for more than we think we have */
	    if(remBytes < size*nitems)	
		nitems = remBytes/size;
	    }
	}
    ret = fread(ptr, size, nitems, stream);
    FixupSamples(ptr, format, ret, READ_FLAG);  /* allow modification */
    return ret;
    }

size_t SFfwrite(ptr, format, nitems, stream)
    void *ptr;
    int  format;
    size_t nitems;
    FILE *stream;
    {
    size_t	size;
    long 	retval;

    FixupSamples(ptr, format, nitems, WRITE_FLAG);  /* allow modification */
    size = SFSampleSizeOf(format);
    retval = fwrite(ptr, size, nitems, stream);
    /* unfix the samples again */
    FixupSamples(ptr, format, nitems, READ_FLAG);
#ifdef DEBUG
    fprintf(stderr, "SFfwrite: wrote %ld of %d, pos=%ld\n", nitems, format, ftell(stream));
#endif /* DEBUG */
    return retval;
    }

int SFfseek(stream, offset, ptrname)
    FILE *stream;
    long offset;
    int  ptrname;
    {
    if(ptrname == SEEK_SET)
	offset += SFHeDaLen(stream);
    return( fseek(stream, offset, ptrname) );
    }

int SFftell(stream)
    FILE *stream;
    {
    return(ftell(stream) - SFHeDaLen(stream));
    }

int SFfeof(stream)
    FILE *stream;
    {
    long pos, end;

    /* actually want to check for run over sample region .. */
    pos = ftell(stream);
    if(pos >= 0)
	{
	end = SFLastSampTell(stream);
	if(end > 0)
	    {
	    if(pos < end)
		return 0;
	    else
		return -1;
	    }
	}
    /* if that doesn't work */
    return(feof(stream));
    }

