/***************************************************************\
*   sndf.h							*
*   header defs for sound files					*
*   'inspired' by the NeXT SND system				*
*   10aug90 dpwe						*
\***************************************************************/

#ifndef _SNDF_H_
#define _SNDF_H_

#include <stdio.h>
#ifdef sun
#include <sys/stdtypes.h>	/* for size_t */
#endif /* sun */

#ifdef THINK_C
extern char sfdir_path[];	/* for buildsfname() - you must provide */
#endif
extern char *programName;	/* for SFDie() - you must provide */

extern char *sndfExtn;		/* defined as file extn for mode e.g. ".irc" */
extern int  SFerror;    	/* internal error flag */

/* these special strings, passed as paths to open functions, use 
   the appropriate streams instead of actual files - dpwe 1993may26 */
#define SF_STDIN_STR "(stdin)"
#define SF_STDOUT_STR "(stdout)"

#define SFDFLTBYTS 4

typedef struct sfstruct
    {
    long    magic;	    /* magic number to identify */
    long    headBsize;	    /* byte offset from start to data */
    long    dataBsize;	    /* number of bytes of data */
    int     format;	    /* format specifier */
    int     channels;	    /* mono/stereo etc  */
    float   samplingRate;   /* .. of the data   */
    char    info[SFDFLTBYTS];	/* extendable byte area */
    } SFSTRUCT;

/***********************/
/* Codes for structure */
/***********************/

#define SFMAGIC 	107365L
#define SF_UNK_LEN	-1L		/* when dataBsize is unavailable */

#define SF_BYTE_SIZE_MASK	15	/* SFMT && SIZE_MASK == # bytes */
#define _ONEBYTE		1
#define _TWOBYTE		2
#define _FOURBYTE		4
#define _EIGHTBYTE		8
#define _LINEAR			0
#define	_FLOAT			32
#define _ALAW			64
#define _ULAW			128
#define _OFFSET			256
/* so.. */
#define SFMT_CHAR  	(_ONEBYTE+_LINEAR)
#define SFMT_CHAR_OFFS  	(_ONEBYTE+_LINEAR+_OFFSET)
#define SFMT_ALAW  	(_ONEBYTE+_ALAW)
#define SFMT_ULAW  	(_ONEBYTE+_ULAW)
#define SFMT_SHORT 	(_TWOBYTE+_LINEAR)
#define SFMT_LONG  	(_FOURBYTE+_LINEAR)
#define SFMT_FLOAT 	(_FOURBYTE+_FLOAT)
#define SFMT_DOUBLE  	(_EIGHTBYTE+_FLOAT)

/* just convenient symbols for the channels field */
#define SF_MONO		1
#define SF_STEREO	2
#define SF_QUAD		4

/* Error codes returned by PVOC file functions */
#define SFE_UNK_ERR	1	/* unknown (other) error (any number other than below) */
#define SFE_OK	    0	/* no error*/
#define SFE_NOPEN   -1	/* couldn't open file */
#define SFE_NSF     -2	/* not a sound file */
#define SFE_MALLOC  -3	/* couldn't allocate memory */
#define SFE_RDERR   -4	/* read error */
#define SFE_WRERR   -5	/* write error */

/*** yukky ANSI bodge (please tell me the right way!) ****/
#ifndef FLOATARG
#ifdef __STDC__
#define FLOATARG double		/* you can't have floats in prototypes! */
#else
#define FLOATARG double
#endif /* def __STDC__ */
#endif /* ndef FLOATARG */

/****************************************/
/**** Function prototypes for export ****/
/****************************************/

#ifdef __STDC__

int  SFInfoSize(SFSTRUCT *snd);
     /* Return number of info bytes in header */

char *SFDataLoc(SFSTRUCT *snd); 
     /* Return pointer to data from header */

/* int  SFSampleSizeOf(int format); */ /* defined in sndfutil.c now */
     /* return number of bytes per sample in given fmt */

int SFSampleFrameSize(SFSTRUCT *snd);	
     /* return number of bytes per sample frame in snd */

int  SFAlloc(SFSTRUCT **psnd, long dataBsize, int format, FLOATARG srate,
	     int chans, int infoBsize);
     /* Allocate memory for a new SFSTRUCT & data block; fill in header
        Returns SFE_MALLOC (& **psnd = NULL) if malloc fails    */

void SFFree(SFSTRUCT *snd);	
     /* destroy in-memory sound */

FILE *SFOpenRdHdr(char *path, SFSTRUCT *snd, int infoBmax);
     /* Opens named file as a stream, reads & checks header, returns stream
        Attempts to read up to infoBmax info bytes, at least 4.
        No error codes; returns NULL on fail (SFE_NOPEN, _NSF, _RDERR) */

int  SFRdXInfo(FILE *file, SFSTRUCT *snd, int done);
     /* fetch any extra info bytes needed in view of newly-read header
        Passed open file handle, snd includes destination for new info,
        done = infobytes already read, other info read from snd */

FILE *SFOpenAllocRdHdr(char *path, SFSTRUCT **psnd);
     /* find sizeof header, alloc space for it, read it in.
        returns NULL on fail (SFE_NOPEN, _NSF, _RDERR) */

int  SFAllocRdHdr(FILE *file, SFSTRUCT **psnd);	
    /* allocate & read a header from open file */

FILE *SFOpenWrHdr(char *path, SFSTRUCT *snd);
     /*  Creates file in path & writes out header, returns stream for access
         No error codes, just returns NULL on fail 
         (SFE_NSF, _NOPEN, _WRERR)	*/

void SFCloseWrHdr(FILE *file, SFSTRUCT *snd);			
     /* After opening via SFOpenWrHdr and writing sound to the stream, may
        need to update header with a postieri data (sound length, more info).
        SFCloseWrHdr works 3 ways:
	- snd is NULL	 ->   look at file ptr, rewrite length to hdr
	- snd exists, but dataBlen is 0
			 ->   fill in new dataBlen from file ptr, write hdr
	- snd exists including dataBlen
			 ->   rewrite whole header
        Ignores any fails & just closes file as best it can.
	file MUST BE READWRITE (mode "w+" or "a+")
	IF MODE JUST "w", THIS silently TRASHES THE HEADER on ultrix 4.0 ! */

int  SFReadHdr(FILE *file, SFSTRUCT *snd, int infoBmax);
/* sample the header of an opened file, at most infoBmax info bytes */

int  SFWriteHdr(FILE *file, SFSTRUCT *snd, int infoBmax);
/*  Utility to write out header to open file
    infoBmax optionally limits the number of info chars to write.
    infoBmax = 0 => write out all bytes implied by .headBsize field
    infoBmax >= SFDFLTBYS => write out only infoBmax infobytes */

int  SFRead(char *path, SFSTRUCT **psnd);
     /* Opens & checks file specified in path.  Allocs for & reads whole sound.
        Poss errors: SFE_NOPEN, SFE_NSF, SFE_MALLOC, SFE_RDERR */

int  SFWrite(char *path, SFSTRUCT *snd);
     /* Write out the soundfile to the file named in path.
        Poss errors: SFE_NSF, SFE_NOPEN, SFE_WRERR */

char *SFErrMsg(int err);	
     /* convert error code to message */
void  SFDie(int err, char *msg);	/* exit routine */

char *buildsfname(char *name);		/*  build a soundfile name (SFDIR) */

size_t SFfread(void *ptr, int type, size_t nitems, FILE *stream);
size_t SFfwrite(void *ptr, int type, size_t nitems, FILE *stream);
int SFfseek(FILE *stream, long offset, int ptrname);
int SFftell(FILE *stream);
int SFfeof(FILE *stream);

#else /* no argument prototypes */

int  SFInfoSize();
char *SFDataLoc(); 
int  SFSampleSizeOf();
int  SFSampleFrameSize();
int  SFAlloc();
void SFFree();	
FILE *SFOpenRdHdr();
FILE *SFOpenAllocRdHdr();
int  SFRdXInfo();
int  SFAllocRdHdr();
FILE *SFOpenWrHdr();
void SFCloseWrHdr();			
int  SFRead();
int  SFWrite();
char *SFErrMsg();	
void  SFDie();
char *buildsfname();
size_t SFfread();
size_t SFfwrite();
int SFfseek();
int SFftell();
int SFfeof();

#endif /* def __STDC__ */

#endif /* _SNDF_H_ */
